'''`bestdori.utils`

杂项模块'''
from typing import Literal, Any, Dict, Tuple

from httpx import Response

from .utils import API, RES,ASSETS
from .network import Api, Res, Assets
from bestdori.exceptions import (
    AssetsNotExistError
)

# 将十六进制颜色代码转换为 RGB 元组
def hexto_rgb(hex: str) -> Tuple[int, int, int]:
    '''将十六进制颜色代码转换为 RGB 元组

    参数:
        hex (str): 十六进制颜色代码

    返回:
        Tuple[int, int, int]: RGB 元组
    '''
    if hex[0] == '#':
        hex = hex[1:]
    if len(hex) == 3:
        hex = hex[0] * 2 + hex[1] * 2 + hex[2] * 2
    rgb = tuple(int(hex[i:i+2], 16) for i in (0, 2, 4))
    if len(rgb) != 3:
        raise ValueError('十六进制颜色代码')
    return rgb

# 获取称号资源
def get_degree(
    degree_name: str,
    server: Literal['jp', 'en', 'tw', 'cn', 'kr']
) -> bytes:
    '''获取称号资源

    参数:
        degree_name (str): 称号名称
        
        server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器
            `jp`: 日服
            `en`: 英服
            `tw`: 台服
            `cn`: 国服
            `kr`: 韩服

    返回:
        bytes: 称号资源字节数据 `bytes`
    '''
    try:
        return Assets(ASSETS['thumb']['degree'].format(degree_name=degree_name), server).get()
    except AssetsNotExistError:
        raise AssetsNotExistError(f'称号 {degree_name}-{server}')

# 异步获取称号资源
async def get_degree_async(
    degree_name: str,
    server: Literal['jp', 'en', 'tw', 'cn', 'kr']
) -> bytes:
    '''获取称号资源

    参数:
        degree_name (str): 称号名称
        
        server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器
            `jp`: 日服
            `en`: 英服
            `tw`: 台服
            `cn`: 国服
            `kr`: 韩服

    返回:
        bytes: 称号资源字节数据 `bytes`
    '''
    try:
        return await Assets(ASSETS['thumb']['degree'].format(degree_name=degree_name), server).aget()
    except AssetsNotExistError:
        raise AssetsNotExistError(f'称号 {degree_name}-{server}')

# 获取总技能信息
def get_skill_all(index: Literal[10]=10) -> Dict[str, Dict[str, Any]]:
    '''获取总技能信息

    参数:
        index (Literal[10], optional): 指定获取哪种 `all.json`
            `10`: 获取所有已有技能的信息 `all.5.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总技能信息
    '''
    return Api(API['all']['skills'].format(index=index)).get().json()

# 异步获取总技能信息
async def get_skill_all_async(index: Literal[10]=10) -> Dict[str, Dict[str, Any]]:
    '''获取总技能信息

    参数:
        index (Literal[10], optional): 指定获取哪种 `all.json`
            `10`: 获取所有已有技能的信息 `all.5.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总技能信息
    '''
    response = await Api(API['all']['skills'].format(index=index)).aget()
    if isinstance(response, Response): return response.json()
    return await response.json()

# 获取总乐队信息
def get_bands_all(index: Literal[1]=1) -> Dict[str, Dict[str, Any]]:
    '''获取总乐队信息

    参数:
        index (Literal[1], optional): 指定获取哪种 `all.json`
            `1`: 获取所有已有乐队的名称信息 `all.1.json`，默认为该项

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总乐队信息
    '''
    return Api(API['bands']['all'].format(index=index)).get().json()

# 异步获取总乐队信息
async def get_bands_all_async(index: Literal[1]=1) -> Dict[str, Dict[str, Any]]:
    '''获取总乐队信息

    参数:
        index (Literal[1], optional): 指定获取哪种 `all.json`
            `1`: 获取所有已有乐队的名称信息 `all.1.json`，默认为该项

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总乐队信息
    '''
    response = await Api(API['bands']['all'].format(index=index)).aget()
    if isinstance(response, Response): return response.json()
    return await response.json()

# 获取主要乐队信息
def get_bands_main(index: Literal[1]=1) -> Dict[str, Dict[str, Any]]:
    '''获取主要乐队信息

    参数:
        index (Literal[1], optional): 指定获取哪种 `main.json`
            `1`: 获取所有主要乐队的名称信息 `main.1.json`，默认为该项

    返回:
        Dict[str, Dict[str, Any]]: 获取到的主要乐队信息
    '''
    return Api(API['bands']['main'].format(index=index)).get().json()

# 异步获取主要乐队信息
async def get_bands_main_async(index: Literal[1]=1) -> Dict[str, Dict[str, Any]]:
    '''获取主要乐队信息

    参数:
        index (Literal[1], optional): 指定获取哪种 `main.json`
            `1`: 获取所有主要乐队的名称信息 `main.1.json`，默认为该项

    返回:
        Dict[str, Dict[str, Any]]: 获取到的主要乐队信息
    '''
    response = await Api(API['bands']['main'].format(index=index)).aget()
    if isinstance(response, Response): return response.json()
    return await response.json()

# 获取乐队 logo
def get_band_logo(
    id: int,
    type: Literal['logoS', 'logoL', 'logoL_Mask'],
    server: Literal['jp', 'en', 'tw', 'cn', 'kr']
) -> bytes:
    '''获取乐队 logo

    参数:
        id (int): 乐队 ID
        
        type (Literal[&#39;logoS&#39;, &#39;logoL&#39;, &#39;logoL_Mask&#39;]): logo 类型
            `logoS`: 小 logo
            `logoL`: 大 logo
            `logoL_Mask`: 大 logo 遮罩
        
        server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

    返回:
        bytes: 乐队 logo 字节数据 `bytes`
    '''
    return Assets(ASSETS['band']['logo'].format(id=id, type=type), server).get()

# 异步获取乐队 logo
async def get_band_logo_async(
    id: int,
    type: Literal['logoS', 'logoL', 'logoL_Mask'],
    server: Literal['jp', 'en', 'tw', 'cn', 'kr']
) -> bytes:
    '''获取乐队 logo

    参数:
        id (int): 乐队 ID
        
        type (Literal[&#39;logoS&#39;, &#39;logoL&#39;, &#39;logoL_Mask&#39;]): logo 类型
            `logoS`: 小 logo
            `logoL`: 大 logo
            `logoL_Mask`: 大 logo 遮罩
        
        server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

    返回:
        bytes: 乐队 logo 字节数据 `bytes`
    '''
    return await Assets(ASSETS['band']['logo'].format(id=id, type=type), server).aget()

# 获取乐队图标
def get_band_icon(id: str) -> bytes:
    '''获取乐队图标

    参数:
        id (str): 乐队 ID

    返回:
        bytes: 乐队图标字节数据
    '''
    return Res(RES['icon']['svg'].format(name=f'band_{id}')).get()

# 异步获取乐队图标
async def get_band_icon_async(id: str) -> bytes:
    '''获取乐队图标

    参数:
        id (str): 乐队 ID

    返回:
        bytes: 乐队图标字节数据
    '''
    return await Res(RES['icon']['svg'].format(name=f'band_{id}')).aget()

# 获取服务器图标
def get_server_icon(server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
    '''获取服务器图标

    参数:
        server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 服务器名称
            `jp`: 日服
            `en`: 英服
            `tw`: 台服
            `cn`: 国服
            `kr`: 韩服

    返回:
        bytes: 服务器图标字节数据
    '''
    return Res(RES['icon']['svg'].format(name=server)).get()

# 异步获取服务器图标
async def get_server_icon_async(server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
    '''获取服务器图标

    参数:
        server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 服务器名称
            `jp`: 日服
            `en`: 英服
            `tw`: 台服
            `cn`: 国服
            `kr`: 韩服

    返回:
        bytes: 服务器图标字节数据
    '''
    return await Res(RES['icon']['svg'].format(name=server)).aget()



'''`bestdori.cards`

BanG Dream! 卡牌相关操作'''
from typing import Any, Dict, Literal

from aiohttp import ClientResponseError
from httpx import Response, HTTPStatusError

from .utils.utils import API, RES, ASSETS
from .post import get_list, get_list_async
from .utils.network import Api, Res, Assets
from .exceptions import (
    NoDataException,
    CardNotExistError
)

# 获取总卡牌信息
def get_all(index: Literal[0, 5]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总卡牌信息

    参数:
        index (Literal[0, 5], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有卡牌 ID `all.0.json`
            `5`: 获取所有已有卡牌的简洁信息 `all.5.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总卡牌信息
    '''
    return Api(API['cards']['all'].format(index=index)).get().json()

# 异步获取总卡牌信息
async def get_all_async(index: Literal[0, 5]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总卡牌信息

    参数:
        index (Literal[0, 5], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有卡牌 ID `all.0.json`
            `5`: 获取所有已有卡牌的简洁信息 `all.5.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总卡牌信息
    '''
    response = await Api(API['cards']['all'].format(index=index)).aget()
    
    if isinstance(response, Response):
        return response.json()
    else:
        return await response.json()

# 获取属性图标
def get_attribute_icon(attribute: Literal['powerful', 'pure', 'cool', 'happy']) -> bytes:
    '''获取属性图标

    参数:
        attribute (Literal[&#39;powerful&#39;, &#39;pure&#39;, &#39;cool&#39;, &#39;happy&#39;]): 属性名称
            `powerful`: POWERFUL
            `pure`: PURE
            `cool`: COOL
            `happy`: HAPPY

    返回:
        bytes: 属性图标字节数据
    '''
    return Res(RES['icon']['svg'].format(name=f'{attribute}')).get()

# 异步获取属性图标
async def get_attribute_icon_async(attribute: Literal['powerful', 'pure', 'cool', 'happy']) -> bytes:
    '''获取属性图标

    参数:
        attribute (Literal[&#39;powerful&#39;, &#39;pure&#39;, &#39;cool&#39;, &#39;happy&#39;]): 属性名称
            `powerful`: POWERFUL
            `pure`: PURE
            `cool`: COOL
            `happy`: HAPPY

    返回:
        bytes: 属性图标字节数据
    '''
    return await Res(RES['icon']['svg'].format(name=f'{attribute}')).aget()

# 获取星星图标
def get_star_icon(star: Literal['star', 'star_trained']) -> bytes:
    '''获取星星图标

    参数:
        star (Literal[&#39;star&#39;, &#39;star_trained&#39;]): 星标种类
            `star`: 普通星标
            `star_trained`: 训练后星标

    返回:
        bytes: 星星图标字节数据
    '''
    return Res(RES['icon']['png'].format(name=f'{star}')).get()

# 异步获取星星图标
async def get_star_icon_async(star: Literal['star', 'star_trained']) -> bytes:
    '''获取星星图标

    参数:
        star (Literal[&#39;star&#39;, &#39;star_trained&#39;]): 星标种类
            `star`: 普通星标
            `star_trained`: 训练后星标

    返回:
        bytes: 星星图标字节数据
    '''
    return await Res(RES['icon']['png'].format(name=f'{star}')).aget()

# 获取卡牌完整边框
def get_frame(level: Literal[1, 2, 3, 4, 5]) -> bytes:
    '''获取卡牌完整边框

    参数:
        level (Literal[1, 2, 3, 4, 5]): 边框星级

    返回:
        bytes: 卡牌完整边框字节数据
    '''
    return Res(RES['image']['png'].format(name=f'frame-{level}')).get()

# 异步获取卡牌完整边框
async def get_frame_async(level: Literal[1, 2, 3, 4, 5]) -> bytes:
    '''获取卡牌完整边框

    参数:
        level (Literal[1, 2, 3, 4, 5]): 边框星级

    返回:
        bytes: 卡牌完整边框字节数据
    '''
    return await Res(RES['image']['png'].format(name=f'frame-{level}')).aget()

# 获取卡牌缩略图边框
def get_card_frame(level: Literal[1, 2, 3, 4, 5]) -> bytes:
    '''获取卡牌缩略图边框

    参数:
        level (Literal[1, 2, 3, 4, 5]): 边框星级

    返回:
        bytes: 卡牌缩略图边框字节数据
    '''
    return Res(RES['image']['png'].format(name=f'card-{level}')).get()

# 异步获取卡牌缩略图边框
async def get_card_frame_async(level: Literal[1, 2, 3, 4, 5]) -> bytes:
    '''获取卡牌缩略图边框

    参数:
        level (Literal[1, 2, 3, 4, 5]): 边框星级

    返回:
        bytes: 卡牌缩略图边框字节数据
    '''
    return await Res(RES['image']['png'].format(name=f'card-{level}')).aget()

# 卡牌类
class Card:
    '''卡牌类

    参数:
        id (int): 卡牌 ID
    '''
    # 初始化
    def __init__(self, id: int) -> None:
        '''卡牌类

        参数:
            id (int): 卡牌 ID
        '''
        self.id: int = id
        '''卡牌 ID'''
        self.__info: Dict[str, Any] = {}
        '''卡牌信息存储'''
        return
    
    # 卡牌标题
    @property
    def prefix(self) -> str:
        '''卡牌标题'''
        info = self.__info
        # 获取 prefix 数据
        if (prefix := info.get('prefix', None)) is None:
            raise NoDataException('卡牌标题')
        # 获取第一个非 None 卡牌标题
        try:
            return next(filter(lambda x: x is not None, prefix))
        except StopIteration:
            raise NoDataException('卡牌标题')
    
    # 卡牌所在默认服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''卡牌所在默认服务器'''
        info = self.__info
        # 获取 releasedAt 数据
        if (released_at := info.get('releasedAt', None)) is None:
            raise NoDataException('卡牌发布时间')
        # 根据 releasedAt 数据判断服务器
        if released_at[0] is not None: return 'jp'
        elif released_at[1] is not None: return 'en'
        elif released_at[2] is not None: return 'tw'
        elif released_at[3] is not None: return 'cn'
        elif released_at[4] is not None: return 'kr'
        else:
            raise NoDataException('卡牌所在服务器')
    
    # 获取卡牌信息
    def get_info(self) -> Dict[str, Any]:
        '''获取卡牌信息

        返回:
            Dict[str, Any]: 卡牌详细信息
        '''
        try:
            response = Api(
                API['cards']['info'].format(id=self.id)
            ).get()
        except HTTPStatusError as exception:
            if exception.response.status_code == 404:
                raise CardNotExistError(self.id)
            else:
                raise exception
        
        self.__info = dict(response.json())
        return self.__info
    
    # 异步获取卡牌信息
    async def get_info_async(self) -> Dict[str, Any]:
        '''获取卡牌信息

        返回:
            Dict[str, Any]: 卡牌详细信息
        '''
        try:
            response = await Api(
                API['cards']['info'].format(id=self.id)
            ).aget()
        except HTTPStatusError as exception:
            if exception.response.status_code == 404:
                raise CardNotExistError(self.id)
            else:
                raise exception
        except ClientResponseError as exception:
            if exception.status == 404:
                raise CardNotExistError(self.id)
            else:
                raise exception
        
        if isinstance(response, Response):
            self.__info = dict(response.json())
        else:
            self.__info = dict(await response.json())
        return self.__info
    
    # 获取缓存信息
    def __get_info_cache(self) -> Dict[str, Any]:
        '''获取缓存信息

        返回:
            Dict[str, Any]: 缓存信息
        '''
        if not self.__info:
            return self.get_info()
        return self.__info
    
    # 异步获取缓存信息
    async def __get_info_cache_async(self) -> Dict[str, Any]:
        '''获取缓存信息

        返回:
            Dict[str, Any]: 缓存信息
        '''
        if not self.__info:
            return await self.get_info_async()
        return self.__info
    
    # 获取卡牌评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> Dict[str, Any]:
        '''获取卡牌评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            Dict[str, Any]: 搜索结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的评论总数
                    "posts": ... # List[Dict[str, Any]] 列举出的评论
                }
                ```
        '''
        return get_list(
            category_name='CARD_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 异步获取卡牌评论
    async def get_comment_async(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> Dict[str, Any]:
        '''获取卡牌评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            Dict[str, Any]: 搜索结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的评论总数
                    "posts": ... # List[Dict[str, Any]] 列举出的评论
                }
                ```
        '''
        return await get_list_async(
            category_name='CARD_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 获取卡牌完整图片
    def get_card(self, type: Literal['normal', 'after_training']) -> bytes:
        '''获取卡牌完整图片

        参数:
            type (Literal[&#39;normal&#39;, &#39;after_training&#39;]): 指定特训前或特训后

        返回:
            bytes: 卡牌完整图片字节数据 `bytes`
        '''
        # 获取卡牌数据包名称
        info = self.__get_info_cache()
        if (resource_set_name := info.get('resourceSetName', None)) is None:
            raise ValueError('无法获取卡牌数据包名称。')
        return Assets(
            ASSETS['characters']['resourceset'].format(
                resource_set_name=resource_set_name, name='card', type=type
            ), self.server
        ).get()
    
    # 异步获取卡牌完整图片
    async def get_card_async(self, type: Literal['normal', 'after_training']) -> bytes:
        '''获取卡牌完整图片

        参数:
            type (Literal[&#39;normal&#39;, &#39;after_training&#39;]): 指定特训前或特训后

        返回:
            bytes: 卡牌完整图片字节数据 `bytes`
        '''
        # 获取卡牌数据包名称
        info = await self.__get_info_cache_async()
        if (resource_set_name := info.get('resourceSetName', None)) is None:
            raise ValueError('无法获取卡牌数据包名称。')
        return await Assets(
            ASSETS['characters']['resourceset'].format(
                resource_set_name=resource_set_name, name='card', type=type
            ), self.server
        ).aget()
    
    # 获取卡牌无背景图片
    def get_trim(self, type: Literal['normal', 'after_training']) -> bytes:
        '''获取卡牌无背景图片

        参数:
            type (Literal[&#39;normal&#39;, &#39;after_training&#39;]): 指定特训前或特训后

        返回:
            bytes: 卡牌无背景图片字节数据 `bytes`
        '''
        # 获取卡牌数据包名称
        info = self.__get_info_cache()
        if (resource_set_name := info.get('resourceSetName', None)) is None:
            raise ValueError('无法获取卡牌数据包名称。')
        return Assets(
            ASSETS['characters']['resourceset'].format(
                resource_set_name=resource_set_name, name='trim', type=type
            ), self.server
        ).get()
    
    # 异步获取卡牌无背景图片
    async def get_trim_async(self, type: Literal['normal', 'after_training']) -> bytes:
        '''获取卡牌无背景图片

        参数:
            type (Literal[&#39;normal&#39;, &#39;after_training&#39;]): 指定特训前或特训后

        返回:
            bytes: 卡牌无背景图片字节数据 `bytes`
        '''
        # 获取卡牌数据包名称
        info = await self.__get_info_cache_async()
        if (resource_set_name := info.get('resourceSetName', None)) is None:
            raise ValueError('无法获取卡牌数据包名称。')
        return await Assets(
            ASSETS['characters']['resourceset'].format(
                resource_set_name=resource_set_name, name='trim', type=type
            ), self.server
        ).aget()
    
    # 获取卡牌缩略图图片
    def get_thumb(self, type: Literal['normal', 'after_training']) -> bytes:
        '''获取卡牌缩略图图片

        参数:
            type (Literal[&#39;normal&#39;, &#39;after_training&#39;]): 指定特训前或特训后

        返回:
            bytes: 卡牌缩略图图片字节数据 `bytes`
        '''
        # 获取卡牌数据包名称
        info = self.__get_info_cache()
        if (resource_set_name := info.get('resourceSetName', None)) is None:
            raise ValueError('无法获取卡牌数据包名称。')
        return Assets(
            ASSETS['thumb']['chara'].format(
                id=str(int(self.id) // 50), resource_set_name=resource_set_name, type=type
            ), self.server
        ).get()
    
    # 异步获取卡牌缩略图图片
    async def get_thumb_async(self, type: Literal['normal', 'after_training']) -> bytes:
        '''获取卡牌缩略图图片

        参数:
            type (Literal[&#39;normal&#39;, &#39;after_training&#39;]): 指定特训前或特训后

        返回:
            bytes: 卡牌缩略图图片字节数据 `bytes`
        '''
        # 获取卡牌数据包名称
        info = await self.__get_info_cache_async()
        if (resource_set_name := info.get('resourceSetName', None)) is None:
            raise ValueError('无法获取卡牌数据包名称。')
        return await Assets(
            ASSETS['thumb']['chara'].format(
                id=str(int(self.id) // 50), resource_set_name=resource_set_name, type=type
            ), self.server
        ).aget()
    
    # 获取 LIVE 服装图片
    def get_livesd(self) -> bytes:
        '''获取 LIVE 服装图片

        返回:
            bytes: 卡牌 LIVE 服装图片字节数据 `bytes`
        '''
        # 获取卡牌 livesd 数据包名称
        info = self.__get_info_cache()
        if (sd_resource_name := info.get('sdResourceName', None)) is None:
            raise ValueError('无法获取卡牌 livesd 数据包名称。')
        return Assets(
            ASSETS['characters']['livesd'].format(
                sd_resource_name=sd_resource_name
            ), self.server
        ).get()
    
    # 异步获取 LIVE 服装图片
    async def get_livesd_async(self) -> bytes:
        '''获取 LIVE 服装图片

        返回:
            bytes: 卡牌 LIVE 服装图片字节数据 `bytes`
        '''
        # 获取卡牌 livesd 数据包名称
        info = await self.__get_info_cache_async()
        if (sd_resource_name := info.get('sdResourceName', None)) is None:
            raise ValueError('无法获取卡牌 livesd 数据包名称。')
        return await Assets(
            ASSETS['characters']['livesd'].format(
                sd_resource_name=sd_resource_name
            ), self.server
        ).aget()



'''`bestdori.miracleticket`

BanG Dream! 自选券相关操作'''
from typing import Any, Dict, List, Literal

from httpx import Response

from .utils.utils import API
from .utils.network import Api
from .exceptions import (
    NoDataException,
    ServerNotAvailableError,
    MiracleTicketExchangeNotExistError
)

# 获取总自选券信息
def get_all(index: Literal[5]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总自选券信息

    参数:
        index (Literal[5], optional): 指定获取哪种 `all.json`
            `5`: 获取所有已有自选券信息 `all.5.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总自选券信息
    '''
    return Api(
        API['all']['miracleTicketExchanges'].format(index=index)
    ).get().json()

# 异步获取总自选券信息
async def get_all_async(index: Literal[5]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总自选券信息

    参数:
        index (Literal[5], optional): 指定获取哪种 `all.json`
            `5`: 获取所有已有自选券信息 `all.5.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总自选券信息
    '''
    response = await Api(
        API['all']['miracleTicketExchanges'].format(index=index)
    ).aget()
    if isinstance(response, Response): return response.json()
    return await response.json()

# 自选券类
class MiracleTicketExchange:
    '''自选券类

    参数:
        id (int): 自选券 ID
    '''
    # 初始化
    def __init__(self, id: int) -> None:
        '''自选券类

        参数:
            id (int): 自选券 ID
        '''
        self.id: int = id
        '''自选券 ID'''
        self.__info: Dict[str, Any] = {}
        '''自选券信息'''
        return
    
    # 自选券标题
    @property
    def name(self) -> str:
        '''自选券标题'''
        info = self.__info
        # 获取 eventName 数据
        if (name := info.get('name', None)) is None:
            raise NoDataException('自选券标题')
        # 获取第一个非 None 自选券标题
        try:
            return next(filter(lambda x: x is not None, name))
        except StopIteration:
            raise NoDataException('自选券标题')
    
    # 自选券默认服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''自选券默认服务器'''
        info = self.__info
        # 获取 ids 数据
        if (ids := info.get('ids', None)) is None:
            raise NoDataException('自选券 ID 列表')
        # 根据 ids 数据判断服务器
        if ids[0] is not None: return 'jp'
        elif ids[1] is not None: return 'en'
        elif ids[2] is not None: return 'tw'
        elif ids[3] is not None: return 'cn'
        elif ids[4] is not None: return 'kr'
        else:
            raise NoDataException('自选券所在服务器')
    
    # 获取自选券信息
    def get_info(self) -> Dict[str, Any]:
        '''获取自选券信息

        返回:
            Dict[str, Any]: 自选券详细信息
        '''
        _all = get_all(5)
        if not self.id in _all.keys():
            raise MiracleTicketExchangeNotExistError(self.id)
        self.__info = _all[str(self.id)]
        return self.__info
    
    # 异步获取自选券信息
    async def get_info_async(self) -> Dict[str, Any]:
        '''获取自选券信息

        返回:
            Dict[str, Any]: 自选券详细信息
        '''
        _all = await get_all_async(5)
        if not self.id in _all.keys():
            raise MiracleTicketExchangeNotExistError(self.id)
        self.__info = _all[str(self.id)]
        return self.__info
    
    # 获取缓存信息
    def __get_info_cache(self) -> Dict[str, Any]:
        '''获取缓存信息

        返回:
            Dict[str, Any]: 缓存信息
        '''
        if not self.__info:
            return self.get_info()
        return self.__info
    
    # 异步获取缓存信息
    async def __get_info_cache_async(self) -> Dict[str, Any]:
        '''获取缓存信息

        返回:
            Dict[str, Any]: 缓存信息
        '''
        if not self.__info:
            return await self.get_info_async()
        return self.__info
    
    # 获取自选券 ID 列表
    def get_ids(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> List[int]:
        '''获取自选券 ID 列表

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            List[int]: 自选券 ID 列表
        '''
        info = self.__get_info_cache()
        # 获取 ids 数据
        if (ids := info.get('ids', None)) is None:
            raise NoDataException('自选券 ID 列表')
        # 判断服务器
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if ids[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return ids[index]
    
    # 异步获取自选券 ID 列表
    async def get_ids_async(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> List[int]:
        '''获取自选券 ID 列表

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            List[int]: 自选券 ID 列表
        '''
        info = await self.__get_info_cache_async()
        # 获取 ids 数据
        if (ids := info.get('ids', None)) is None:
            raise NoDataException('自选券 ID 列表')
        # 判断服务器
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if ids[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return ids[index]



'''`bestdori.events`

BanG Dream! 活动相关操作'''
from typing import Any, Dict, List, Literal, Optional

from aiohttp import ClientResponseError
from httpx import Response, HTTPStatusError

from .utils.utils import API, ASSETS
from .utils.network import Api, Assets
from .eventarchives import EventArchive
from .post import get_list, get_list_async
from .festival import (
    Stage,
    RotationMusic,
    get_stages,
    get_stages_async,
    get_rotation_musics,
    get_rotation_musics_async
)
from .exceptions import (
    NoDataException,
    EventNotExistError,
    AssetsNotExistError,
    EventHasNoStampError,
    ServerNotAvailableError
)

# 获取总活动信息
def get_all(index: Literal[0, 5, 6]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总活动信息

    参数:
        index (Literal[0, 5, 6], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有活动 ID `all.0.json`
            `5`: 获取所有已有活动的简洁信息 `all.5.json`
            `6`: 获取所有已有活动的简洁信息 `all.6.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总活动信息
    '''
    return Api(API['events']['all'].format(index=index)).get().json()

# 异步获取总活动信息
async def get_all_async(index: Literal[0, 5, 6]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总活动信息

    参数:
        index (Literal[0, 5, 6], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有活动 ID `all.0.json`
            `5`: 获取所有已有活动的简洁信息 `all.5.json`
            `6`: 获取所有已有活动的简洁信息 `all.6.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总活动信息
    '''
    response = await Api(API['events']['all'].format(index=index)).aget()
    if isinstance(response, Response):
        return response.json()
    else:
        return await response.json()

# 活动类
class Event:
    '''活动类

    参数:
        id (int): 活动 ID
    '''
    # 初始化
    def __init__(self, id: int) -> None:
        '''活动类

        参数:
            id (int): 活动 ID
        '''
        self.id: int = id
        '''活动 ID'''
        self.archive: EventArchive = EventArchive(self.id)
        '''活动数据'''
        self.__info: Dict[str, Any] = {}
        '''活动信息'''
        return
    
    # 活动标题
    @property
    def name(self) -> str:
        '''活动标题'''
        info = self.__info
        # 获取 eventName 数据
        if (event_name := info.get('eventName', None)) is None:
            raise NoDataException('活动标题')
        # 获取第一个非 None 活动标题
        try:
            return next(filter(lambda x: x is not None, event_name))
        except StopIteration:
            raise NoDataException('活动标题')
    
    # 活动默认服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''活动默认服务器

        返回:
            Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]: 歌曲所在服务器
        '''
        info = self.__info
        # 获取 startAt 数据
        if (start_at := info.get('startAt', None)) is None:
            raise NoDataException('活动起始时间')
        # 根据 startAt 数据判断服务器
        if start_at[0] is not None: return 'jp'
        elif start_at[1] is not None: return 'en'
        elif start_at[2] is not None: return 'tw'
        elif start_at[3] is not None: return 'cn'
        elif start_at[4] is not None: return 'kr'
        else:
            raise NoDataException('活动所在服务器')
    
    # 获取活动信息
    def get_info(self) -> Dict[str, Any]:
        '''获取活动信息

        返回:
            Dict[str, Any]: 活动详细信息
        '''
        try:
            response = Api(
                API['events']['info'].format(id=self.id)
            ).get()
        except HTTPStatusError as exception:
            if exception.response.status_code == 404:
                raise EventNotExistError(self.id)
            else:
                raise exception
        
        self._info = dict(response.json())
        return self._info
    
    # 异步获取活动信息
    async def get_info_async(self) -> Dict[str, Any]:
        '''获取活动信息

        返回:
            Dict[str, Any]: 活动详细信息
        '''
        try:
            response = await Api(
                API['events']['info'].format(id=self.id)
            ).aget()
        except HTTPStatusError as exception:
            if exception.response.status_code == 404:
                raise EventNotExistError(self.id)
            else:
                raise exception
        except ClientResponseError as exception:
            if exception.status == 404:
                raise EventNotExistError(self.id)
            else:
                raise exception
        
        if isinstance(response, Response):
            self._info = dict(response.json())
        else:
            self._info = dict(await response.json())
        return self._info
    
    # 获取缓存信息
    def __get_info_cache(self) -> Dict[str, Any]:
        '''获取缓存信息

        返回:
            Dict[str, Any]: 活动详细信息
        '''
        if not self._info:
            return self.get_info()
        return self._info
    
    # 异步获取缓存信息
    async def __get_info_cache_async(self) -> Dict[str, Any]:
        '''获取缓存信息

        返回:
            Dict[str, Any]: 活动详细信息
        '''
        if not self._info:
            return await self.get_info_async()
        return self._info
    
    # 获取活动评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> Dict[str, Any]:
        '''获取活动评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            Dict[str, Any]: 搜索结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的评论总数
                    "posts": ... # List[Dict[str, Any]] 列举出的评论
                }
                ```
        '''
        return get_list(
            category_name='EVENT_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 异步获取活动评论
    async def get_comment_async(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> Dict[str, Any]:
        '''获取活动评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            Dict[str, Any]: 搜索结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的评论总数
                    "posts": ... # List[Dict[str, Any]] 列举出的评论
                }
                ```
        '''
        return await get_list_async(
            category_name='EVENT_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 获取活动缩略图图像
    def get_banner(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
        '''获取活动缩略图图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            bytes: 活动缩略图图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = self.__get_info_cache()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return Assets(
            ASSETS['event']['banner'].format(
                asset_bundle_name=asset_bundle_name
            ), server
        ).get()
    
    # 异步获取活动缩略图图像
    async def get_banner_async(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
        '''获取活动缩略图图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            bytes: 活动缩略图图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = await self.__get_info_cache_async()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return await Assets(
            ASSETS['event']['banner'].format(
                asset_bundle_name=asset_bundle_name
            ), server
        ).aget()
    
    # 获取活动 logo 图像
    def get_logo(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
        '''获取活动 logo 图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            bytes: 活动 logo 图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = self.__get_info_cache()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return Assets(
            ASSETS['event']['logo'].format(
                asset_bundle_name=asset_bundle_name
            ), server
        ).get()
    
    # 异步获取活动 logo 图像
    async def get_logo_async(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
        '''获取活动 logo 图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            bytes: 活动 logo 图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = await self.__get_info_cache_async()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return await Assets(
            ASSETS['event']['logo'].format(
                asset_bundle_name=asset_bundle_name
            ), server
        ).aget()

    # 获取活动主界面图像
    def get_topscreen(
        self,
        server: Literal['jp', 'en', 'tw', 'cn', 'kr'],
        type: Literal['bg', 'trim']
    ) -> bytes:
        '''获取活动主界面图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器
            
            type (Literal[&#39;bg&#39;, &#39;trim&#39;]): 图像类型
                `bg`: 背景图像
                `trim`: 角色图像

        返回:
            bytes: 活动主界面图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = self.__get_info_cache()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return Assets(
            ASSETS['event']['topscreen'].format(
                asset_bundle_name=asset_bundle_name, type=type
            ), server
        ).get()
    
    # 异步获取活动主界面图像
    async def get_topscreen_async(
        self,
        server: Literal['jp', 'en', 'tw', 'cn', 'kr'],
        type: Literal['bg', 'trim']
    ) -> bytes:
        '''获取活动主界面图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器
            
            type (Literal[&#39;bg&#39;, &#39;trim&#39;]): 图像类型
                `bg`: 背景图像
                `trim`: 角色图像

        返回:
            bytes: 活动主界面图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = await self.__get_info_cache_async()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return await Assets(
            ASSETS['event']['topscreen'].format(
                asset_bundle_name=asset_bundle_name, type=type
            ), server
        ).aget()
    
    # 获取活动奖励稀有表情
    def get_stamp(self) -> bytes:
        '''获取活动奖励稀有表情

        返回:
            bytes: 活动奖励稀有表情字节数据 `bytes`
        '''
        info = self.__get_info_cache()
        # 获取活动点数奖励列表
        if (point_rewards := info.get('pointRewards', None)) is None:
            raise NoDataException('活动奖励')
        # 获取第一个非 None 奖励列表
        try:
            point_reward = next(filter(lambda x: x is not None, point_rewards))
        except StopIteration:
            raise NoDataException('活动奖励')
        # 获取 rewardType 为 stamp 的活动奖励
        try:
            reward = next(filter(lambda x: x['rewardType'] == 'stamp', point_reward))
        except StopIteration:
            raise EventHasNoStampError(self.id)
        stamp_id = reward['rewardId']
        # 获取全部贴纸资源
        stamps = Api(
            API['all']['stamps'].format(index=2)
        ).get().json()
        if (stamp := stamps.get(stamp_id, None)) is None:
            raise AssetsNotExistError(f'贴纸 {stamp_id}')
        # 获取贴纸资源
        image_name = stamp['imageName']
        return Assets(
            ASSETS['stamp']['get'].format(image_name=image_name), self.server
        ).get()
    
    # 异步获取活动奖励稀有表情
    async def get_stamp_async(self) -> bytes:
        '''获取活动奖励稀有表情

        返回:
            bytes: 活动奖励稀有表情字节数据 `bytes`
        '''
        info = await self.__get_info_cache_async()
        # 获取活动点数奖励列表
        if (point_rewards := info.get('pointRewards', None)) is None:
            raise NoDataException('活动奖励')
        # 获取第一个非 None 奖励列表
        try:
            point_reward = next(filter(lambda x: x is not None, point_rewards))
        except StopIteration:
            raise NoDataException('活动奖励')
        # 获取 rewardType 为 stamp 的活动奖励
        try:
            reward = next(filter(lambda x: x['rewardType'] == 'stamp', point_reward))
        except StopIteration:
            raise EventHasNoStampError(self.id)
        stamp_id = reward['rewardId']
        # 获取全部贴纸资源
        stamps = await Api(
            API['all']['stamps'].format(index=2)
        ).aget()
        if isinstance(stamps, Response):
            stamps = stamps.json()
        else:
            stamps = await stamps.json()
        if (stamp := stamps.get(stamp_id, None)) is None:
            raise AssetsNotExistError(f'贴纸 {stamp_id}')
        # 获取贴纸资源
        image_name = stamp['imageName']
        return await Assets(
            ASSETS['stamp']['get'].format(image_name=image_name), self.server
        ).aget()

    # 获取排名分数线
    def get_top(
        self,
        server: Optional[Literal[0, 1, 2, 3, 4]]=None,
        mid: Literal['0']='0',
        latest: Literal['1']='1'
    ) -> Dict[str, List[Dict[str, Any]]]:
        '''获取排名分数线

        参数:
            server (Optional[Literal[0, 1, 2, 3, 4]], optional): 指定服务器
                `0`: 日服
                `1`: 英服
                `2`: 台服
                `3`: 国服
                `4`: 韩服
            mid (Literal[&#39;0&#39;], optional): 指定是否为中间分数线，默认为 `0`
            latest (Literal[&#39;1&#39;], optional): 指定是否为最终分数线，默认为 `1`

        返回:
            Dict[str, List[Dict[str, Any]]]: 排名分数线数据
        '''
        if server is None:
            self.__get_info_cache()
            if self.server == 'jp': server = 0
            elif self.server == 'en': server = 1
            elif self.server == 'tw': server = 2
            elif self.server == 'cn': server = 3
            elif self.server == 'kr': server = 4
            else: raise NoDataException('活动服务器')
        return self.archive.get_top(server, mid, latest)
    
    # 异步获取排名分数线
    async def get_top_async(
        self,
        server: Optional[Literal[0, 1, 2, 3, 4]]=None,
        mid: Literal['0']='0',
        latest: Literal['1']='1'
    ) -> Dict[str, List[Dict[str, Any]]]:
        '''获取排名分数线

        参数:
            server (Optional[Literal[0, 1, 2, 3, 4]], optional): 指定服务器
                `0`: 日服
                `1`: 英服
                `2`: 台服
                `3`: 国服
                `4`: 韩服
            mid (Literal[&#39;0&#39;], optional): 指定是否为中间分数线，默认为 `0`
            latest (Literal[&#39;1&#39;], optional): 指定是否为最终分数线，默认为 `1`

        返回:
            Dict[str, List[Dict[str, Any]]]: 排名分数线数据
        '''
        if server is None:
            await self.__get_info_cache_async()
            if self.server == 'jp': server = 0
            elif self.server == 'en': server = 1
            elif self.server == 'tw': server = 2
            elif self.server == 'cn': server = 3
            elif self.server == 'kr': server = 4
            else: raise NoDataException('活动服务器')
        return await self.archive.get_top_async(server, mid, latest)

    # 获取团队 LIVE 佳节活动歌曲循环数据
    def get_rotation_musics(self) -> List[RotationMusic]:
        '''获取团队 LIVE 佳节活动歌曲循环数据

        返回:
            List[Dict[str, Any]]: 团队 LIVE 佳节活动歌曲循环数据
        '''
        info = self.__get_info_cache()
        if (_event_type := info.get('eventTyppe', '')) != 'festival':
            raise ValueError(f'Rotation musics are only available for festival events, not {_event_type}.')
        return get_rotation_musics(self.id)
    
    # 异步获取团队 LIVE 佳节活动歌曲循环数据
    async def get_rotation_musics_async(self) -> List[RotationMusic]:
        '''获取团队 LIVE 佳节活动歌曲循环数据

        返回:
            List[Dict[str, Any]]: 团队 LIVE 佳节活动歌曲循环数据
        '''
        info = await self.__get_info_cache_async()
        if (_event_type := info.get('eventTyppe', '')) != 'festival':
            raise ValueError(f'Rotation musics are only available for festival events, not {_event_type}.')
        return await get_rotation_musics_async(self.id)
    
    # 获取团队 LIVE 佳节活动舞台数据
    def get_stages(self) -> List[Stage]:
        '''获取团队 LIVE 佳节活动舞台数据

        返回:
            List[Dict[str, Any]]: 团队 LIVE 佳节活动舞台数据
        '''
        info = self.__get_info_cache()
        if (_event_type := info.get('eventTyppe', '')) != 'festival':
            raise ValueError(f'Stages are only available for festival events, not {_event_type}.')
        return get_stages(self.id)
    
    # 异步获取团队 LIVE 佳节活动舞台数据
    async def get_stages_async(self) -> List[Stage]:
        '''获取团队 LIVE 佳节活动舞台数据

        返回:
            List[Dict[str, Any]]: 团队 LIVE 佳节活动舞台数据
        '''
        info = await self.__get_info_cache_async()
        if (_event_type := info.get('eventTyppe', '')) != 'festival':
            raise ValueError(f'Stages are only available for festival events, not {_event_type}.')
        return await get_stages_async(self.id)



'''`bestdori.ayachan`

ayachan 的各种 API 调用整合'''
from pathlib import Path
from mimetypes import guess_type
from typing import Any, Dict, Union, Literal

from httpx import Response

from .utils import API, Api
from bestdori.charts import Chart
from .exceptions import SonolusException

DIFF_STR = ['easy', 'normal', 'hard', 'expert', 'special']

# 自定义谱面分析
def chart_metrics_custom(chart: Chart, diff: Literal[0, 1, 2, 3, 4]) -> Dict[str, Any]:
    '''自定义谱面分析

    参数:
        chart (Chart): 谱面
        diff (Literal[0, 1, 2, 3, 4]): 难度类型

    返回:
        Dict[str, Any]: 分析结果
    '''
    return Api(API['chart_metrics']['custom'].format(diff_str=DIFF_STR[diff])).post(data=chart.to_list()).json()

# 异步自定义谱面分析
async def chart_metrics_custom_async(chart: Chart, diff: Literal[0, 1, 2, 3, 4]) -> Dict[str, Any]:
    '''自定义谱面分析

    参数:
        chart (Chart): 谱面
        diff (Literal[0, 1, 2, 3, 4]): 难度类型

    返回:
        Dict[str, Any]: 分析结果
    '''
    response = await Api(API['chart_metrics']['custom'].format(diff_str=DIFF_STR[diff])).apost(data=chart.to_list())
    if isinstance(response, Response): return response.json()
    return await response.json()

# BanG Dream 谱面分析
def chart_metrics_bandori(
    id: int,
    diff: Literal[0, 1, 2, 3, 4]
) -> Dict[str, Any]:
    '''BanG Dream 谱面分析

    参数:
        id (int): 歌曲 ID
        diff (Literal[0, 1, 2, 3, 4]): 难度类型

    返回:
        Dict[str, Any]: 分析结果
    '''
    return Api(API['chart_metrics']['bandori'].format(chart_id=id, diff_str=DIFF_STR[diff])).get().json()

# 异步 BanG Dream 谱面分析
async def chart_metrics_bandori_async(
    id: int,
    diff: Literal[0, 1, 2, 3, 4]
) -> Dict[str, Any]:
    '''BanG Dream 谱面分析

    参数:
        id (int): 歌曲 ID
        diff (Literal[0, 1, 2, 3, 4]): 难度类型

    返回:
        Dict[str, Any]: 分析结果
    '''
    response = await Api(API['chart_metrics']['bandori'].format(chart_id=id, diff_str=DIFF_STR[diff])).aget()
    if isinstance(response, Response): return response.json()
    return await response.json()

# Bestdori 谱面分析
def chart_metrics_bestdori(id: int) -> Dict[str, Any]:
    '''Bestdori 谱面分析

    参数:
        id (int): 谱面 ID

    返回:
        Dict[str, Any]: 分析结果
    '''
    return Api(API['chart_metrics']['bestdori'].format(chart_id=id)).get().json()

# 异步 Bestdori 谱面分析
async def chart_metrics_bestdori_async(id: int) -> Dict[str, Any]:
    '''Bestdori 谱面分析

    参数:
        id (int): 谱面 ID

    返回:
        Dict[str, Any]: 分析结果
    '''
    response = await Api(API['chart_metrics']['bestdori'].format(chart_id=id)).aget()
    if isinstance(response, Response): return response.json()
    return await response.json()

# Sonolus 谱面测试上传
def post_sonolus_levels(
    title: str,
    bgm: Union[str, Path],
    difficulty: int,
    hidden: bool,
    lifetime: Literal[3600, 10800, 21600, 43200, 86400],
    chart: Chart
) -> int:
    '''Sonolus 谱面测试

    参数:
        title (str): 谱面标题
        bgm (Union[str, Path]): 音乐文件
        difficulty (int): 谱面难度
        hidden (bool): 谱面隐藏
        lifetime (Literal[3600, 10800, 21600, 43200, 86400]): 存活时间
        chart (Chart): 上传谱面

    返回:
        int: 测试服 ID
    '''
    # 转换文件路径并获取名称
    if isinstance(bgm, str):
        bgm = Path(bgm)
    bgm_name = bgm.name
    
    # 构建数据
    data: Dict[str, Any] = {
        'title': title,
        'difficulty': str(difficulty),
        'lifetime': str(lifetime),
        'chart': chart.json()
    }
    if hidden:
        data['hidden'] = True
    
    # 准备文件
    file = open(bgm, 'rb')
    files = {'bgm': (bgm_name, file, guess_type(bgm)[0])}
    
    # 发送请求
    response = Api(API['levels']).post(data=data, files=files)
    file.close()
    if (uid := response.json().get('uid', None)) is None:
        description = response.json().get('description', None)
        detail = response.json().get('detail', None)
        raise SonolusException(f'{description} {detail}')
    return uid

# 异步 Sonolus 谱面测试上传
async def post_sonolus_levels_async(
    title: str,
    bgm: Union[str, Path],
    difficulty: int,
    hidden: bool,
    lifetime: Literal[3600, 10800, 21600, 43200, 86400],
    chart: Chart
) -> int:
    '''Sonolus 谱面测试

    参数:
        title (str): 谱面标题
        bgm (Union[str, Path]): 音乐文件
        difficulty (int): 谱面难度
        hidden (bool): 谱面隐藏
        lifetime (Literal[3600, 10800, 21600, 43200, 86400]): 存活时间
        chart (Chart): 上传谱面

    返回:
        int: 测试服 ID
    '''
    # 转换文件路径并获取名称
    if isinstance(bgm, str):
        bgm = Path(bgm)
    bgm_name = bgm.name
    
    # 构建数据
    data: Dict[str, Any] = {
        'title': title,
        'difficulty': str(difficulty),
        'lifetime': str(lifetime),
        'chart': chart.json()
    }
    if hidden:
        data['hidden'] = True
    
    # 准备文件
    file = open(bgm, 'rb')
    files = {'bgm': (bgm_name, file, guess_type(bgm)[0])}
    
    # 发送请求
    response = await Api(API['levels']).apost(data=data, files=files)
    file.close()
    
    if isinstance(response, Response): response_json = response.json()
    else: response_json = await response.json()
    
    if (uid := response_json.get('uid', None)) is None:
        description = response_json.get('description', None)
        detail = response_json.get('detail', None)
        raise SonolusException(f'{description} {detail}')
    return uid

# Sonolus 测试服谱面获取
def get_sonolus_levels(uid: int) -> Chart:
    '''Sonolus 测试服谱面获取

    参数:
        uid (int): 测试服 ID

    返回:
        Chart: 谱面
    '''
    response = Api(API['levels']['get'].format(uid=uid)).get()
    return Chart.normalize(response.json())

# 异步 Sonolus 测试服谱面获取
async def get_sonolus_levels_async(uid: int) -> Chart:
    '''Sonolus 测试服谱面获取

    参数:
        uid (int): 测试服 ID

    返回:
        Chart: 谱面
    '''
    response = await Api(API['levels']['get'].format(uid=uid)).aget()
    if isinstance(response, Response): return Chart.normalize(response.json())
    return Chart.normalize(await response.json())

# Sonolus 测试服谱面信息获取
def sonolus_levels(uid: int) -> Dict[str, Any]:
    '''Sonolus 测试服谱面信息获取

    参数:
        uid (int): 测试服 ID

    返回:
        Dict[str, Any]: 谱面信息
    '''
    return Api(API['levels']['info'].format(uid=uid)).get().json()

# 异步 Sonolus 测试服谱面信息获取
async def sonolus_levels_async(uid: int) -> Dict[str, Any]:
    '''Sonolus 测试服谱面信息获取

    参数:
        uid (int): 测试服 ID

    返回:
        Dict[str, Any]: 谱面信息
    '''
    response = await Api(API['levels']['info'].format(uid=uid)).aget()
    if isinstance(response, Response): return response.json()
    return await response.json()

from .utils import settings as settings



'''`bestdori.models` 对象定义模块'''

from .content import Br as Br
from .content import Link as Link
from .content import Text as Text
from .content import Emoji as Emoji
from .content import Image as Image
from .content import Content as Content
from .content import Heading as Heading
from .content import Mention as Mention
from .content import ContentList as ContentList

from .note import BPM as BPM
from .note import Slide as Slide
from .note import Single as Single
from .note import NoteType as NoteType
from .note import Connection as Connection
from .note import Directional as Directional



'''`bestdori.ayachan.utils.utils`

ayachan 工具库模块'''


API = {
    'version': 'v2/version',
    'chart_metrics': {
        'bandori': 'v2/chart/metrics/bandori/{chart_id}/{diff_str}',
        'bestdori': 'v2/chart/metrics/bestdori/{chart_id}',
        'custom': 'v2/chart/metrics/custom/{diff_str}'
    },
    'levels': {
        'post': 'https://sonolus.ayachan.fun/test/sonolus/levels',
        'info': 'https://sonolus.ayachan.fun/test/sonolus/levels/{uid}',
        'get': 'https://sonolus.ayachan.fun/test/sonolus/levels/{uid}/bdv2.json'
    }
}
'''ayachan API 集合'''



'''`bestdori.user`

BanG Dream! 歌曲相关操作'''
from http.cookies import SimpleCookie
from typing import Any, Dict, Literal, Optional

from httpx import Response

from . import settings
from .utils.utils import API
from .utils.network import Api
from .post import get_list, get_list_async

# 用户类
class User:
    '''用户类

    参数:
        username (str): 用户名
    '''
    # 初始化
    def __init__(self, username: str) -> None:
        '''用户类

        参数:
            username (str): 用户名
        '''        
        self.username: str = username
        '''用户名'''
        self.__info: Dict[str, Any] = {}
        '''用户信息'''
        return
    
    # 获取用户详细信息
    def get_info(self) -> Dict[str, Any]:
        '''获取用户详细信息

        返回:
            Dict[str, Any]: 用户详细信息
        '''
        response = Api(
            API['user']['info']
        ).get(params={'username': self.username})
        self.__info = dict(response.json())
        return self.__info
    
    # 异步获取用户详细信息
    async def get_info_async(self) -> Dict[str, Any]:
        '''获取用户详细信息

        返回:
            Dict[str, Any]: 用户详细信息
        '''
        response = await Api(
            API['user']['info']
        ).aget(params={'username': self.username})
        if isinstance(response, Response):
            self.__info = dict(response.json())
        else:
            self.__info = dict(await response.json())
        return self.__info
    
    # 获取用户帖子
    def get_posts(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> Dict[str, Any]:
        '''获取用户帖子

        参数:
            limit (int, optional): 展示出的帖子数，默认 20
            offset (int, optional): 忽略前面的 `offset` 个帖子，默认 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 帖子排序，默认时间倒序

        返回:
            Dict[str, Any]: 获取结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的帖子总数
                    "posts": ... # List[Dict[str, Any]] 列举出的帖子
                }
                ```
        '''
        return get_list(
            username=self.username, limit=limit, offset=offset, order=order
        )
    
    # 异步获取用户帖子
    async def get_posts_async(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> Dict[str, Any]:
        '''获取用户帖子

        参数:
            limit (int, optional): 展示出的帖子数，默认 20
            offset (int, optional): 忽略前面的 `offset` 个帖子，默认 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 帖子排序，默认时间倒序

        返回:
            Dict[str, Any]: 获取结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的帖子总数
                    "posts": ... # List[Dict[str, Any]] 列举出的帖子
                }
                ```
        '''
        return await get_list_async(
            username=self.username, limit=limit, offset=offset, order=order
        )
    
    # 获取用户谱面
    def get_charts(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> Dict[str, Any]:
        '''获取用户谱面

        参数:
            limit (int, optional): 展示出的谱面数，默认 20
            offset (int, optional): 忽略前面的 `offset` 个谱面，默认 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 谱面排序，默认时间倒序

        返回:
            Dict[str, Any]: 获取结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的谱面总数
                    "posts": ... # List[Dict[str, Any]] 列举出的谱面
                }
                ```
        '''
        return get_list(
            username=self.username,
            category_name='SELF_POST',
            category_id='chart',
            limit=limit,
            offset=offset,
            order=order
        )
    
    # 异步获取用户谱面
    async def get_charts_async(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> Dict[str, Any]:
        '''获取用户谱面

        参数:
            limit (int, optional): 展示出的谱面数，默认 20
            offset (int, optional): 忽略前面的 `offset` 个谱面，默认 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 谱面排序，默认时间倒序

        返回:
            Dict[str, Any]: 获取结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的谱面总数
                    "posts": ... # List[Dict[str, Any]] 列举出的谱面
                }
                ```
        '''
        return await get_list_async(
            username=self.username,
            category_name='SELF_POST',
            category_id='chart',
            limit=limit,
            offset=offset,
            order=order
        )
    
    # 获取用户文本帖子
    def get_texts(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> Dict[str, Any]:
        '''获取用户文本帖子

        参数:
            limit (int, optional): 展示出的帖子数，默认 20
            offset (int, optional): 忽略前面的 `offset` 个帖子，默认 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 帖子排序，默认时间倒序

        返回:
            Dict[str, Any]: 获取结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的帖子总数
                    "posts": ... # List[Dict[str, Any]] 列举出的帖子
                }
                ```
        '''
        return get_list(
            username=self.username,
            category_name='SELF_POST',
            category_id='text',
            limit=limit,
            offset=offset,
            order=order
        )
    
    # 异步获取用户文本帖子
    async def get_texts_async(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> Dict[str, Any]:
        '''获取用户文本帖子

        参数:
            limit (int, optional): 展示出的帖子数，默认 20
            offset (int, optional): 忽略前面的 `offset` 个帖子，默认 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 帖子排序，默认时间倒序

        返回:
            Dict[str, Any]: 获取结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的帖子总数
                    "posts": ... # List[Dict[str, Any]] 列举出的帖子
                }
                ```
        '''
        return await get_list_async(
            username=self.username,
            category_name='SELF_POST',
            category_id='text',
            limit=limit,
            offset=offset,
            order=order
        )
    
    # 获取用户故事
    def get_storys(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> Dict[str, Any]:
        '''获取用户故事

        参数:
            limit (int, optional): 展示出的故事数，默认 20
            offset (int, optional): 忽略前面的 `offset` 个故事，默认 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 故事排序，默认时间倒序

        返回:
            Dict[str, Any]: 获取结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的故事总数
                    "posts": ... # List[Dict[str, Any]] 列举出的故事
                }
                ```
        '''
        return get_list(
            username=self.username,
            category_name='SELF_POST',
            category_id='story',
            limit=limit,
            offset=offset,
            order=order
        )
    
    # 异步获取用户故事
    async def get_storys_async(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> Dict[str, Any]:
        '''获取用户故事

        参数:
            limit (int, optional): 展示出的故事数，默认 20
            offset (int, optional): 忽略前面的 `offset` 个故事，默认 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 故事排序，默认时间倒序

        返回:
            Dict[str, Any]: 获取结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的故事总数
                    "posts": ... # List[Dict[str, Any]] 列举出的故事
                }
                ```
        '''
        return await get_list_async(
            username=self.username,
            category_name='SELF_POST',
            category_id='story',
            limit=limit,
            offset=offset,
            order=order
        )

# 自身用户类
class Me(User):
    '''自身用户类

    参数:
        username (str): 用户名
        password (str): 密码
    '''
    # 初始化
    def __init__(self, username: str, password: str) -> None:
        '''自身用户类

        参数:
            username (str): 用户名
            password (str): 密码
        '''
        super().__init__(username)
        self.password: str = password
        '''密码'''
        self.__cookies: Optional[SimpleCookie] = None
        '''用户 Cookies'''
        self.__me: Dict[str, Any] = {}
        '''用户自我信息'''
        return
    
    # 用户登录
    @classmethod
    def login(cls, username: str, password: str) -> 'Me':
        '''用户登录

        参数:
            username (str): 用户名
            password (str): 密码

        返回:
            Me: 自身用户对象
        '''
        me = cls(username, password)
        response = Api(
            API['user']['login']
        ).post(data={'username': me.username, 'password': me.password})
        me.__cookies = SimpleCookie(response.cookies)
        me.me()
        settings._me = me
        return me
    
    # 用户异步登录
    @classmethod
    async def alogin(cls, username: str, password: str) -> 'Me':
        '''用户登录

        参数:
            username (str): 用户名
            password (str): 密码

        返回:
            Me: 自身用户对象
        '''
        me = cls(username, password)
        response = await Api(
            API['user']['login']
        ).apost(data={'username': me.username, 'password': me.password})
        if isinstance(response, Response):
            me.__cookies = SimpleCookie(response.cookies)
        else:
            me.__cookies = response.cookies
        await me.ame()
        settings._me = me
        return me
    
    # 获取用户 Cookies
    @property
    def cookies(self) -> SimpleCookie:
        '''获取用户 Cookies'''
        if self.__cookies is None:
            raise ValueError(f'用户 {self.username} 未登录。')
        return self.__cookies
    
    # 获取用户自我信息
    def me(self) -> Dict[str, Any]:
        '''获取用户自我信息

        返回:
            Dict[str, Any]: 自我信息
        '''
        if len(self.__me) == 0:
            response = Api(API['user']['me']).get(cookies=self.cookies)
            self.__me = dict(response.json())
        return self.__me
    
    # 异步获取用户自我信息
    async def ame(self) -> Dict[str, Any]:
        '''获取用户自我信息

        返回:
            Dict[str, Any]: 自我信息
        '''
        if len(self.__me) == 0:
            response = await Api(API['user']['me']).aget(cookies=self.cookies)
            if isinstance(response, Response):
                self.__me = dict(response.json())
            else:
                self.__me = dict(await response.json())
        return self.__me



from setuptools import find_packages, setup

with open('README.md', 'r') as readme:
    long_description = readme.read()

setup(
    name='bestdori-api',
    version='1.1.2',
    author='WindowsSov8',
    author_email='qwertyuiop2333@hotmail.com',
    description='Bestdori 的各种 API 调用整合，另外附带部分功能',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/WindowsSov8forUs/bestdori-api',
    include_package_data=False,
    packages=find_packages(),
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.8',
    install_requires=[
        'httpx>=0.22.0',
        'aiohttp>=3.8.1',
        'typing_extensions>=4.5.0'
    ]
)


'''`bestdori.player`

BanG Dream! 玩家信息相关操作'''

from typing import Any, Dict, Optional

from httpx import Response

from .utils.utils import API
from .utils.network import Api
from .exceptions import PlayerNotExistError

class Player:
    '''玩家类

    参数:
        id (int): 玩家 ID
        server (str): 服务器
    '''
    def __init__(self, id: int, server: str) -> None:
        self.id: int = id
        self.server: str = server
        self.__info: Optional[Dict[str, Any]] = None
    
    def get_info(self) -> Dict[str, Any]:
        '''获取玩家信息

        返回:
            Dict[str, Any]: 玩家信息
        '''
        params = {
            'mode': 2
        }
        _info = Api(
            API['player']['info'].format(server=self.server, id=self.id)
        ).get(params=params).json()
        if not _info.get('result', False):
            raise ValueError(f'Invalid Server: {self.server}')
        _data: Dict[str, Any] = _info.get('data', {})
        _profile: Optional[Dict[str, Any]] = _data.get('profile', None)
        if _profile is None:
            raise PlayerNotExistError(self.server, self.id)
        self.__info = _profile
        
        return self.__info
    
    async def get_info_async(self) -> Dict[str, Any]:
        '''获取玩家信息

        返回:
            Dict[str, Any]: 玩家信息
        '''
        params = {
            'mode': 2
        }
        response = await Api(
            API['player']['info'].format(server=self.server, id=self.id)
        ).aget(params=params)
        
        if isinstance(response, Response):
            _info = response.json()
        else:
            _info = await response.json()
        
        if not _info.get('result', False):
            raise ValueError(f'Invalid Server: {self.server}')
        _data: Dict[str, Any] = _info.get('data', {})
        _profile: Optional[Dict[str, Any]] = _data.get('profile', None)
        if _profile is None:
            raise PlayerNotExistError(self.server, self.id)
        self.__info = _profile
        
        return self.__info



'''`bestdori.songs`

BanG Dream! 歌曲相关操作'''
from typing import Any, Dict, List, Literal

from aiohttp import ClientResponseError
from httpx import Response, HTTPStatusError

from .charts import Chart
from .utils.utils import API, ASSETS
from .utils.network import Api, Assets
from .post import get_list, get_list_async
from .exceptions import (
    NoDataException,
    DiffNotExistError,
    SongNotExistError
)

# 获取总歌曲信息
def get_all(index: Literal[0, 5, 7]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总歌曲信息

    参数:
        index (Literal[0, 5], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有歌曲 ID `all.0.json`
            `5`: 获取所有已有歌曲的简洁信息 `all.5.json`，默认为该项
            `7`: 获取所有已有歌曲的较为详细信息 `all.7.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总歌曲信息
    '''
    return Api(API['songs']['all'].format(index=index)).get().json()

# 异步获取总歌曲信息
async def get_all_async(index: Literal[0, 5, 7]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总歌曲信息

    参数:
        index (Literal[0, 5], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有歌曲 ID `all.0.json`
            `5`: 获取所有已有歌曲的简洁信息 `all.5.json`，默认为该项
            `7`: 获取所有已有歌曲的较为详细信息 `all.7.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总歌曲信息
    '''
    response = await Api(API['songs']['all'].format(index=index)).aget()
    if isinstance(response, Response):
        return response.json()
    else:
        return await response.json()

# 歌曲封面内部类
class Jacket:
    '''歌曲封面类

    参数:
        url (str): 封面链接
        bytes (bytes): 封面字节数据
    '''
    # 初始化
    def __init__(
        self,
        index: int,
        jacket_image: str,
        server: Literal['jp', 'en', 'tw', 'cn', 'kr']
    ) -> None:
        '''歌曲封面类'''
        self._index: int = index
        '''数据包序列号'''
        self._jacket_image: str = jacket_image
        '''封面文件名'''
        self._server: Literal['jp', 'en', 'tw', 'cn', 'kr'] = server
        '''封面所在服务器'''
        return
    
    # 封面 url
    @property
    def url(self) -> str:
        '''封面 url'''
        return Assets(
            ASSETS['songs']['musicjacket'].format(
                index=self._index, jacket_image=self._jacket_image
            ), self._server
        ).get_url()
    
    # 获取封面字节数据
    def get_bytes(self) -> bytes:
        '''获取封面字节数据'''
        return Assets(
            ASSETS['songs']['musicjacket'].format(
                index=self._index, jacket_image=self._jacket_image
            ), self._server
        ).get()
    
    # 异步获取封面字节数据
    async def get_bytes_async(self) -> bytes:
        '''获取封面字节数据'''
        return await Assets(
            ASSETS['songs']['musicjacket'].format(
                index=self._index, jacket_image=self._jacket_image
            ), self._server
        ).aget()

# 歌曲类
class Song:
    '''歌曲类

    参数:
        id (int): 歌曲 ID
    '''
    # 初始化
    def __init__(self, id: int) -> None:
        '''歌曲类

        参数:
            id (int): 歌曲 ID
        '''
        self.id: int = id
        '''歌曲 ID'''
        self.__info: Dict[str, Any] = {}
        '''歌曲信息'''
        return
    
    # 获取歌曲信息
    def get_info(self) -> Dict[str, Any]:
        '''获取歌曲信息

        返回:
            Dict[str, Any]: 歌曲详细信息
        '''
        try:
            response = Api(
                API['songs']['info'].format(id=self.id)
            ).get()
        except HTTPStatusError as exception:
            if exception.response.status_code == 404:
                raise SongNotExistError(self.id)
            else:
                raise exception
        
        self.__info = dict(response.json())
        return self.__info
    
    # 异步获取歌曲信息
    async def get_info_async(self) -> Dict[str, Any]:
        '''获取歌曲信息

        返回:
            Dict[str, Any]: 歌曲详细信息
        '''
        try:
            response = await Api(
                API['songs']['info'].format(id=self.id)
            ).aget()
        except HTTPStatusError as exception:
            if exception.response.status_code == 404:
                raise SongNotExistError(self.id)
            else:
                raise exception
        except ClientResponseError as exception:
            if exception.status == 404:
                raise SongNotExistError(self.id)
            else:
                raise exception
        
        if isinstance(response, Response):
            self.__info = dict(response.json())
        else:
            self.__info = dict(await response.json())
        return self.__info
    
    # 歌曲所在默认服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''歌曲所在默认服务器'''
        info = self.__info
        # 获取 publishedAt 数据
        if (published_at := info.get('publishedAt', None)) is None:
            raise NoDataException('歌曲发布时间')
        # 根据 publishedAt 数据判断服务器
        if published_at[0] is not None: return 'jp'
        elif published_at[1] is not None: return 'en'
        elif published_at[2] is not None: return 'tw'
        elif published_at[3] is not None: return 'cn'
        elif published_at[4] is not None: return 'kr'
        else:
            raise NoDataException('歌曲服务器')
    
    # 歌曲名称
    @property
    def name(self) -> str:
        '''歌曲名称'''
        info = self.__info
        # 获取 musicTitle 数据
        if (music_title := info.get('musicTitle', None)) is None:
            raise NoDataException('歌曲名称')
        # 获取第一个非 None 歌曲名称
        try:
            return next(filter(lambda x: x is not None, music_title))
        except StopIteration:
            raise NoDataException('歌曲名称')
    
    # 歌曲封面
    @property
    def jacket(self) -> List[Jacket]:
        '''歌曲封面'''
        # 获取数据包序列号
        quotient, remainder = divmod(self.id, 10)
        if remainder == 0:
            index = self.id
        else:
            index = (quotient + 1) * 10
        
        info = self.__info
        if (jacket_image := info.get('jacketImage', None)) is None:
            raise NoDataException('歌曲封面资源')
        jacket: List[Jacket] = []
        
        for image in jacket_image:
            jacket.append(Jacket(index, image, self.server))
        
        return jacket
    
    # 获取歌曲谱面
    def get_chart(
        self,
        diff: Literal['easy', 'normal', 'hard', 'expert', 'special']='expert'
    ) -> Chart:
        '''获取歌曲谱面

        参数:
            diff (Literal[&#39;easy&#39;, &#39;normal&#39;, &#39;hard&#39;, &#39;expert&#39;, &#39;special&#39;], optional): 难度名称

        返回:
            Chart: 获取到的谱面对象
        '''
        try:
            chart = Chart.get_chart(self.id, diff)
            return chart
        except HTTPStatusError as e:
            if e.response.status_code == 404:
                # 难度不存在
                raise DiffNotExistError(diff)
            else:
                raise e
    
    # 异步获取歌曲谱面
    async def get_chart_async(
        self,
        diff: Literal['easy', 'normal', 'hard', 'expert', 'special']='expert'
    ) -> Chart:
        '''获取歌曲谱面

        参数:
            diff (Literal[&#39;easy&#39;, &#39;normal&#39;, &#39;hard&#39;, &#39;expert&#39;, &#39;special&#39;], optional): 难度名称

        返回:
            Chart: 获取到的谱面对象
        '''
        try:
            chart = await Chart.get_chart_async(self.id, diff)
            return chart
        except HTTPStatusError as e:
            if e.response.status_code == 404:
                # 难度不存在
                raise DiffNotExistError(diff)
            else:
                raise e
        except ClientResponseError as e:
            if e.status == 404:
                # 难度不存在
                raise DiffNotExistError(diff)
            else:
                raise e
    
    # 获取歌曲音频
    def get_bgm(self) -> bytes:
        '''获取歌曲音频

        返回:
            bytes: 歌曲音频字节数据 `bytes`
        '''
        return Assets(ASSETS['songs']['sound'].format(id=self.id), self.server).get()
    
    # 异步获取歌曲音频
    async def get_bgm_async(self) -> bytes:
        '''获取歌曲音频

        返回:
            bytes: 歌曲音频字节数据 `bytes`
        '''
        return await Assets(ASSETS['songs']['sound'].format(id=self.id), self.server).aget()
    
    # 获取歌曲评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> Dict[str, Any]:
        '''获取歌曲评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            Dict[str, Any]: 搜索结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的评论总数
                    "posts": ... # List[Dict[str, Any]] 列举出的评论
                }
                ```
        '''
        return get_list(
            category_name='SONG_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 异步获取歌曲评论
    async def get_comment_async(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> Dict[str, Any]:
        '''获取歌曲评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            Dict[str, Any]: 搜索结果
                ```python
                {
                    "result": ... # bool 是否有响应
                    "count": ... # int 搜索到的评论总数
                    "posts": ... # List[Dict[str, Any]] 列举出的评论
                }
                ```
        '''
        return await get_list_async(
            category_name='SONG_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )



'''`bestdori.songmeta`

BanG Dream! 歌曲 Meta 相关操作'''
from typing import Any, Dict, Literal

from httpx import Response

from .utils.utils import API
from .utils.network import Api

# 获取总歌曲 Meta 信息
def get_all(index: Literal[5]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总歌曲 Meta 信息

    参数:
        index (Literal[5], optional): 指定获取哪种 `all.json`
            `5`: 获取所有已有歌曲 Meta 信息 `all.5.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总歌曲 Meta 信息
    '''
    return Api(API['all']['meta'].format(index=index)).get().json()

# 异步获取总歌曲 Meta 信息
async def get_all_async(index: Literal[5]=5) -> Dict[str, Dict[str, Any]]:
    '''获取总歌曲 Meta 信息

    参数:
        index (Literal[5], optional): 指定获取哪种 `all.json`
            `5`: 获取所有已有歌曲 Meta 信息 `all.5.json`

    返回:
        Dict[str, Dict[str, Any]]: 获取到的总歌曲 Meta 信息
    '''
    response = await Api(API['all']['meta'].format(index=index)).aget()
    if isinstance(response, Response):
        return response.json()
    else:
        return await response.json()



'''`bestdori.charts`

谱面相关操作'''
from json import dumps, loads
from typing import Any, Dict, List, Literal

from httpx import Response

from .models.note import *
from .utils.utils import API
from .utils.network import Api

# 谱面数据类
class Statistics:
    '''谱面数据类

    参数:
        time (float): 谱面时长
        notes (int): 谱面音符总数
        bpm (List[float]): 谱面 BPM 范围
        main_bpm (float): 谱面主 BPM
    '''
    # 初始化
    def __init__(self, time: float, notes: int, bpm: List[float], main_bpm: float) -> None:
        '''谱面数据类

        参数:
            time (float): 谱面时长
            notes (int): 谱面音符总数
            bpm (List[float]): 谱面 BPM 范围
            main_bpm (float): 谱面主 BPM
        '''
        self.time: float = time
        '''谱面时长'''
        self.notes: int = notes
        '''谱面音符总数'''
        self.bpm: List[float] = bpm
        '''谱面 BPM 范围'''
        self.main_bpm: float = main_bpm
        '''谱面主 BPM'''
        return

# 谱面类
class Chart(List[NoteType]):
    '''谱面类，统合针对谱面的一层操作

    参数:
        chart (List[Dict[str, Any]]): 原始谱面代码'''
    # 初始化
    def __init__(self, chart: List[Dict[str, Any]]) -> None:
        '''谱面类，统合针对谱面的一层操作

        参数:
            chart (List[Dict[str, Any]]): 原始谱面代码'''
        super().__init__()
        for note in chart:
            # 遍历分类添加
            if note['type'] in ['Long', 'Slide']:
                self.append(Slide(**note))
            elif note['type'] == 'BPM':
                self.append(BPM(**note))
            elif note['type'] == 'Single':
                self.append(Single(**note))
            elif note['type'] == 'Directional':
                self.append(Directional(**note))
            else:
                # 删除其他音符
                continue
        return
    
    # 检查是否为 SP 谱面
    @property
    def is_sp_rhythm(self) -> bool:
        '''是否为使用了 SP 键的谱面'''
        for note in self:
            if isinstance(note, Slide):
                for connection in note.connections:
                    if connection.hidden:
                        return True
            elif isinstance(note, Directional):
                return True
        return False
    
    # 谱面规范化处理
    @classmethod
    def normalize(cls, chart: List[Dict[str, Any]]) -> 'Chart':
        '''谱面规范化处理

        参数:
            chart (List[Dict[str, Any]]): 待处理谱面

        返回:
            Chart: 处理后谱面
        '''
        normalized_chart: cls = cls(chart)
        # 对谱面进行排序
        normalized_chart.sort(key=lambda x: x.beat)
        # 处理可能出现的 BPM 错位
        if not isinstance(normalized_chart[0], BPM):
            offset: float = -1.0 # 记录 offset 修正
            # 第一位不是 BPM，找寻真正的 BPM 线
            for note in normalized_chart:
                if isinstance(note, BPM):
                    offset = note.beat
                    break
            if offset < 0: # 没有找到 BPM
                raise ValueError('谱面内未找到 BPM 线。')
            # 对谱面节拍进行修正
            for note in normalized_chart:
                note.beat_move(-offset)
                if isinstance(note, Slide):
                    for connection in note.connections:
                        if connection.beat < 0:
                            connection.beat = 0
                        else:
                            break
                else:
                    if note.beat < 0:
                        note.beat = 0
                    else:
                        break
        
        # 处理可能出现的不合法滑条节点
        for note in normalized_chart:
            if not isinstance(note, Slide):
                continue
            index: int = 0
            for connection in note.connections:
                if index < (len(note.connections) - 1):
                    if connection.flick:
                        connection.flick = False
                if 0 < index < (len(note.connections) - 1):
                    if connection.skill:
                        connection.skill = False
                index += 1
        
        # 对谱面节拍进行修正
        if normalized_chart[0].beat != 0:
            offset = normalized_chart[0].beat
            for note in normalized_chart:
                note.beat_move(-offset)
        return normalized_chart
    
    # 谱面数据统计
    def count(self) -> Statistics:
        '''谱面数据统计

        返回:
            Statistics: 统计到的谱面详细数据
        '''
        # 初始化统计数据
        start_beat = 0.0 # 谱面开始 beat 值
        end_beat = 0.0 # 谱面结束 beat 值
        prev_bpm = 120.0 # 上一个 BPM 线的 BPM 值
        prev_bpm_beat = 0.0 # 上一个 BPM 线的 beat 值
        total_notes = 0 # 总物量
        bpm_list: List[Dict[str, float]] = [] # BPM 统计列表，统计所有出现的 BPM 及其有效时间
        
        # 遍历谱面数据
        for note in self:
            # 谱面为一个字典列表，每一个 note 都是一个字典
            if isinstance(note, BPM): # 如果当前是 BPM
                if note.bpm >= 0: # 如果当前 BPM 大于等于 0
                    # 如果不是谱面一开始的 BPM 线且已有 note 被记录（即已出现过有效 bpm ）
                    if note.beat > 0 and total_notes > 0:
                        if prev_bpm_beat <= start_beat: # 如果上一个 BPM 线先于第一个 note
                            prev_bpm_beat = start_beat
                        bpm_duration = (note.beat - prev_bpm_beat) * 60.0 / prev_bpm # 计算持续时间
                        bpm_flag: bool = False # 检测 BPM 表中是否已存在指定 BPM
                        for bpm_dict in bpm_list:
                            if bpm_dict['bpm'] == prev_bpm:
                                bpm_dict['duration'] += bpm_duration
                                bpm_flag = True
                                break
                        if not bpm_flag: # 如果 BPM 未被记录
                            bpm_dict = {
                                'bpm': prev_bpm,
                                'duration': bpm_duration
                            }
                            bpm_list.append(bpm_dict)
                    prev_bpm = note.bpm
                    prev_bpm_beat = note.beat
                continue
            
            if isinstance(note, (Single, Directional)): # 如果当前是单键或方向滑键
                # 记录 beat
                if end_beat < note.beat: # 如果当前 beat 更靠后
                    end_beat = note.beat # 始终记录结束 beat
                if start_beat <= 0 or start_beat > note.beat: # 如果未记录起始 beat 或已记录的并不是起始 beat
                    start_beat = note.beat
                
                total_notes += 1 # 累加一个物量
                continue
            
            if isinstance(note, Slide): # 如果是绿条
                # 绿条将会有一个 `connections` 列表用于记录节点
                for connection in note.connections:
                    if not connection.hidden: # 忽略隐藏节点
                        # 记录 beat
                        if end_beat < connection.beat: # 如果当前 beat 更靠后
                            end_beat = connection.beat # 始终记录结束 beat
                        if start_beat <= 0 or start_beat > connection.beat: # 如果未记录起始 beat 或已记录的并不是起始 beat
                            start_beat = connection.beat
                        
                        total_notes += 1 # 累加一个物量
                continue
                    
        # 当走出遍历后表明谱面已遍历至最后一个 note ，进行最后的处理
        if prev_bpm_beat < end_beat: # 如果最后一个 note 在最后一个 BPM 线之前
            bpm_duration = (end_beat - prev_bpm_beat) * 60.0 / prev_bpm # 计算持续时间
            bpm_flag: bool = False # 检测 BPM 表中是否已存在指定 BPM
            for bpm_dict in bpm_list:
                if bpm_dict['bpm'] == prev_bpm:
                    bpm_dict['duration'] += bpm_duration
                    bpm_flag = True
                    break
            if not bpm_flag: # 如果 BPM 未被记录
                bpm_dict = {
                    'bpm': prev_bpm,
                    'duration': bpm_duration
                }
                bpm_list.append(bpm_dict)
                
        # 遍历 BPM 列表，计算总时长并获取 BPM 数值
        duration = 0.0 # 谱面总持续时长
        bpm_main = 0.0 # 主要 BPM
        bpm_main_dura = 0.0 # 主要 BPM 持续时长
        bpm_min = 2147483647.0 # 最低 BPM
        bpm_max = 0.0 # 最高 BPM
        for bpm_info in bpm_list: # 遍历
            if bpm_info['duration'] > bpm_main_dura: # 如果持续时间更长
                bpm_main_dura = bpm_info['duration']
                bpm_main = bpm_info['bpm']
            if bpm_min > bpm_info['bpm']: # 如果更小
                bpm_min = bpm_info['bpm']
            if bpm_max < bpm_info['bpm']: # 如果更大
                bpm_max = bpm_info['bpm']
            duration += bpm_info['duration'] # 累加持续时长
        
        return Statistics(
            duration,
            total_notes,
            [bpm_min, bpm_max] if bpm_min != bpm_max else [bpm_min],
            bpm_main
        )

    # 转换为字典列表对象
    def to_list(self) -> List[Dict[str, Any]]:
        '''将 `Chart` 谱面转换为 `List[Dict[str, Any]]` 对象'''
        chart_data: List[Dict[str, Any]] = []
        for note in self:
            chart_data.append(note.__dict__)
        return chart_data
    
    # 转换为 json 字符串
    def json(self) -> str:
        '''将 `Chart` 谱面转换为 `json` 字符串'''
        return dumps(self.to_list(), ensure_ascii=False)

    # 通过 json 字符串转换为 Chart 谱面
    @classmethod
    def from_json(cls, data: str) -> 'Chart':
        '''通过 `json` 字符串转换为 `Chart` 谱面

        参数:
            data (str): 谱面 `json` 字符串

        返回:
            Chart: 谱面对象 `bestdori.chart.Chart`
        '''
        return cls(loads(data))
    
    # 获取官方谱面
    @classmethod
    def get_chart(
        cls,
        id: int,
        diff: Literal['easy', 'normal', 'hard', 'expert', 'special']='expert'
    ) -> 'Chart':
        '''获取官方谱面

        参数:
            id (int): 谱面 ID
            diff (Literal[&#39;easy&#39;, &#39;normal&#39;, &#39;hard&#39;, &#39;expert&#39;, &#39;special&#39;], optional): 难度名称

        返回:
            Chart: 获取到的谱面对象 `bestdori.chart.Chart`
        '''
        response = Api(API['charts']['info'].format(id=id, diff=diff)).get()
        return cls.normalize(response.json())
    
    # 异步获取官方谱面
    @classmethod
    async def get_chart_async(
        cls,
        id: int,
        diff: Literal['easy', 'normal', 'hard', 'expert', 'special']='expert'
    ) -> 'Chart':
        '''获取官方谱面

        参数:
            id (int): 谱面 ID
            diff (Literal[&#39;easy&#39;, &#39;normal&#39;, &#39;hard&#39;, &#39;expert&#39;, &#39;special&#39;], optional): 难度名称

        返回:
            Chart: 获取到的谱面对象 `bestdori.chart.Chart`
        '''
        response = await Api(API['charts']['info'].format(id=id, diff=diff)).aget()
        if isinstance(response, Response):
            return cls.normalize(response.json())
        else:
            return cls.normalize(await response.json())



'''`bestdori.utils.note`

谱面音符相关模块'''
from typing import Literal, Any, Dict, List
from typing_extensions import override

# 谱面音符类
class NoteType:
    '''谱面音符类'''
    type: str
    '''音符类型'''
    beat: float
    '''节拍数'''
    # 初始化
    def __init__(self, **values) -> None:
        '''初始化'''
        for key, value in values.items():
            if key == 'type':
                continue
            setattr(self, key, value)
        return
    
    def __getitem__(self, key: str) -> Any:
        '''获取属性'''
        return getattr(self, key)
    
    # 节拍数增减
    def beat_move(self, beat: float) -> None:
        '''节拍数增减

        参数:
            beat (float): 移动的节拍数
        '''
        self.beat += beat

# 滑条节点
class Connection:
    '''滑条节点类

    参数:
        beat (float): 节拍数
        
        lane (int): 轨道数
        
        hidden (bool, optional): 是否隐藏'''
    beat: float
    '''节拍数'''
    lane: int
    '''轨道数'''
    hidden: bool = False
    '''是否隐藏'''
    flick: bool = False
    '''是否为滑键'''
    skill: bool = False
    '''是否为技能键'''
    # 初始化
    def __init__(self, **values) -> None:
        '''初始化'''
        for key, value in values.items():
            setattr(self, key, value)
        return
    
    # 字典化
    @property
    def __dict__(self) -> Dict[str, Any]:
        '''字典化'''
        note = {'beat': self.beat, 'lane': self.lane}
        if self.hidden:
            note['hidden'] = self.hidden
        if self.flick:
            note['flick'] = self.flick
        if self.skill:
            note['skill'] = self.skill
        return note

# BPM
class BPM(NoteType):
    '''BPM

    参数:
        bpm (float): BPM 值
        
        beat (float): 节拍数'''
    type: str = 'BPM'
    '''音符类型'''
    bpm: float
    '''BPM 值'''
    # 字典化
    @property
    def __dict__(self) -> Dict[str, Any]:
        '''字典化'''
        return {
            'bpm': self.bpm,
            'beat': self.beat,
            'type': self.type
        }

# 单键
class Single(NoteType):
    '''单键

    参数:
        beat (float): 节拍数
        
        lane (int): 轨道数
        
        flick (bool, optional): 是否为滑键
        
        skill (bool, optional): 是否为技能键'''
    type: str = 'Single'
    '''音符类型'''
    lane: int
    '''轨道数'''
    flick: bool = False
    '''是否为滑键'''
    skill: bool = False
    '''是否为技能键'''
    # 字典化
    @property
    def __dict__(self) -> Dict[str, Any]:
        '''字典化'''
        note = {
            'beat': self.beat,
            'lane': self.lane,
            'type': self.type
        }
        if self.flick:
            note['flick'] = self.flick
        if self.skill:
            note['skill'] = self.skill
        return note

# 方向滑键
class Directional(NoteType):
    '''方向滑键

    参数:
        beat (float): 节拍数
        
        lane (int): 轨道数
        
        width (int): 滑键宽度
        
        direction (Literal[&#39;Left&#39;, &#39;Right&#39;]): 滑键方向'''
    type: str = 'Directional'
    '''音符类型'''
    lane: int
    '''轨道数'''
    width: int
    '''滑键宽度'''
    direction: Literal['Left', 'Right']
    '''滑键方向'''
    # 字典化
    @property
    def __dict__(self) -> Dict[str, Any]:
        '''字典化'''
        note = {
            'beat': self.beat,
            'lane': self.lane,
            'type': self.type,
            'width': self.width,
            'direction': self.direction
        }
        return note

# 滑条
class Slide(NoteType):
    '''滑条

    参数:
        connections (List[Connection]): 滑键节点'''
    type: str = 'Slide'
    '''音符类型'''
    connections: List[Connection]
    '''滑键节点'''
    # 初始化
    def __init__(self, **values) -> None:
        '''初始化'''
        for key, value in values.items():
            if key == 'type':
                continue
            if key == 'connections':
                connections: List[Connection] = []
                # 提取起始节拍数
                self.beat = value[0]['beat']
                
                for connection in value:
                    connections.append(Connection(**connection))
                value = connections
            setattr(self, key, value)
        return
    
    # 字典化
    @property
    def __dict__(self) -> Dict[str, Any]:
        '''字典化'''
        note: Dict[str, Any] = {
            'type': self.type
        }
        if self.connections:
            note['connections'] = [connection.__dict__ for connection in self.connections]
        return note
    
    # 节拍数增减
    @override
    def beat_move(self, beat: float) -> None:
        for connection in self.connections:
            connection.beat += beat
        self.beat = self.connections[0].beat
        return

__all__ = [
    'NoteType',
    'Connection',
    'BPM',
    'Single',
    'Directional',
    'Slide'
]